#!/usr/bin/env python

# Edit this script to add your team's training code.
# Some functions are *required*, but you can edit most parts of the required functions, remove non-required functions, and add your own functions.

################################################################################
#
# Imported functions and variables
#
################################################################################

# Import functions. These functions are not required. You can change or remove them.
from helper_code import *
import numpy as np, os, sys, joblib
from sklearn.impute import SimpleImputer
from sklearn.ensemble import RandomForestClassifier
from sklearn.decomposition import PCA
from sklearn.preprocessing import StandardScaler

from scipy import stats
from scipy.stats import skew, kurtosis
import numpy as np
import math
import persim
from persim import persistent_entropy
from ripser import ripser

import multiprocessing as mp
import gc

#import matplotlib.pyplot as plt

from get_12ECG_features import get_12ECG_features

# Define the Challenge lead sets. These variables are not required. You can change or remove them.
twelve_leads = ('I', 'II', 'III', 'aVR', 'aVL', 'aVF', 'V1', 'V2', 'V3', 'V4', 'V5', 'V6')
six_leads = ('I', 'II', 'III', 'aVR', 'aVL', 'aVF')
four_leads = ('I', 'II', 'III', 'V2')
three_leads = ('I', 'II', 'V2')
two_leads = ('I', 'II')
lead_sets = (twelve_leads, six_leads, four_leads, three_leads, two_leads)

################################################################################
#
# Training model function
#
################################################################################

# Train your model. This function is *required*. You should edit this function to add your code, but do *not* change the arguments of this function.
def training_code(data_directory, model_directory):
    # Find header and recording files.
    print('Finding header and recording files...')

    header_files, recording_files = find_challenge_files(data_directory)
    num_recordings = len(recording_files)

    if not num_recordings:
        raise Exception('No data was provided.')

    # Create a folder for the model if it does not already exist.
    if not os.path.isdir(model_directory):
        os.mkdir(model_directory)

    # Extract the classes from the dataset.
    print('Extracting classes...')

    classes = set()
    for header_file in header_files:
        header = load_header(header_file)

        classes |= set(get_labels(header))
    if all(is_integer(x) for x in classes):
        classes = sorted(classes, key=lambda x: int(x)) # Sort classes numerically if numbers.
    else:
        classes = sorted(classes) # Sort classes alphanumerically if not numbers.
    num_classes = len(classes)

    # Extract the features and labels from the dataset.
    print('Extracting labels...')
    
    labels = np.zeros((num_recordings, num_classes), dtype=np.bool) # One-hot encoding of classes
    features = list()
    
    NSR_count=0
    NSR_exclude = np.array([])
    SB_count=0
    SB_exclude = np.array([])
    
    #Only work with 5000 NSR and SB
    for i in range(num_recordings):
        header = load_header(header_files[i])
        current_labels = get_labels(header)
        for label in current_labels:
            if label in classes:
                if label =='426783006':
                    if NSR_count<5000:
                        NSR_count +=1
                        j = classes.index(label)
                        labels[i, j] = 1
                    else:
                        NSR_exclude = np.append(NSR_exclude,int(i))
                elif label == '426177001':
                    if SB_count<5000:
                        SB_count +=1
                        j = classes.index(label)
                        labels[i, j] = 1
                    else:
                        SB_exclude = np.append(SB_exclude,int(i))
                else:
                    j = classes.index(label)
                    labels[i, j] = 1
    NSR_exclude = np.append(NSR_exclude,SB_exclude)
    woExcess = [t for t in range(num_recordings) if t not in NSR_exclude]
    labels = labels[woExcess]
    
#    print('cleaning...')
    header_files = [header_files[t] for t in woExcess]
    recording_files = [recording_files[t] for t in woExcess]
    num_recordings = len(recording_files)
        
    # Define parameters for random forest classifier.
    n_estimators = 200     # Number of trees in the forest.
    max_leaf_nodes = 20 #maximum number of leaf nodes in each tree.
    min_samples_split = 10
    min_samples_leaf = 5
    random_state = 0 
    
    top12 = [252, 121, 39, 125, 107, 123, 37, 47, 23, 127, 95, 119, 135, 147, 11, 43, 133, 3, 164, 41, 51, 9, 82, 45, 131, 145, 162, 137, 143, 94, 163, 59, 27, 139, 13, 106, 83, 15, 71, 80, 70, 92, 109, 35, 21, 141, 7, 113, 53, 57, 17, 22, 19, 115, 5, 73, 33, 55, 75, 1, 111, 49, 85, 69, 81, 63, 100, 30, 67, 155, 65, 101, 104, 249, 97, 93, 29, 90, 154, 31, 91, 118, 87, 34, 129, 153, 231, 99, 42, 77, 124, 114, 66, 103, 89, 61, 78, 16, 105, 126]

    top6 =  [174, 47, 23, 39, 11, 37, 75, 45, 41, 59, 43, 3, 9, 73, 71, 51, 15, 70, 27, 22, 21, 13, 34, 35, 19, 17, 57, 33, 7, 65, 86, 5, 153, 85, 68, 53, 69, 63, 67, 54, 55, 171, 4, 1, 29, 31, 42, 32, 20, 84, 49, 46, 18, 66, 40, 6, 61, 25, 72, 30, 28, 76, 64, 16, 52, 58, 99, 77, 150, 169, 60, 151, 44, 12, 168, 26, 36, 74, 24, 48, 62, 14, 56, 10, 38, 0, 50, 159, 8, 2, 157, 158, 156, 81, 83, 82, 170, 79, 80, 152]

    top4 =  [148, 23, 11, 47, 9, 15, 46, 3, 49, 13, 35, 17, 27, 21, 19, 44, 45, 60, 5, 41, 37, 7, 39, 33, 22, 43, 34, 1, 58, 42, 59, 29, 31, 25, 40, 18, 30, 16, 6, 28, 20, 48, 38, 4, 127, 32, 51, 36, 24, 145, 26, 12, 50, 73, 14, 0, 2, 10, 124, 125, 143, 142, 8, 56, 133, 55, 131, 130, 132, 54, 57, 52, 126, 139, 53, 68, 66, 62, 65, 137, 64, 136, 61, 123, 99, 67, 135, 111, 69, 134, 105, 63, 87, 110, 144, 93, 86, 116, 98, 122]

    top3 =  [135, 23, 35, 11, 3, 9, 15, 34, 37, 13, 21, 17, 19, 33, 32, 5, 27, 7, 22, 30, 29, 31, 25, 1, 18, 47, 28, 6, 45, 46, 26, 20, 36, 16, 12, 4, 114, 24, 38, 14, 132, 0, 10, 60, 112, 111, 2, 8, 130, 129, 43, 42, 120, 118, 119, 44, 55, 117, 126, 113, 41, 49, 40, 48, 54, 56, 122, 74, 39, 86, 121, 52, 50, 124, 98, 53, 84, 123, 69, 80, 110, 131, 115, 133, 73, 103, 104, 116, 97, 51, 82, 134, 68, 106, 128, 85, 91, 70, 109, 79]

    top2 =  [122, 11, 23, 3, 15, 9, 25, 13, 17, 21, 19, 22, 7, 33, 5, 18, 34, 16, 4, 1, 6, 20, 24, 12, 32, 14, 0, 95, 47, 2, 10, 119, 93, 8, 101, 117, 71, 69, 92, 116, 98, 99, 68, 83, 106, 30, 80, 107, 105, 41, 81, 31, 29, 35, 39, 104, 77, 59, 94, 75, 40, 73, 118, 36, 28, 57, 53, 38, 27, 67, 108, 56, 97, 96, 51, 37, 43, 60, 26, 109, 90, 113, 70, 85, 42, 74, 72, 66, 79, 91, 65, 82, 45, 121, 84, 111, 62, 78, 63, 115]

  
    pool = mp.Pool(mp.cpu_count())
        
        
    print('Extracting 2-lead features...')    
    
    data = np.zeros((num_recordings, (12*2)+2+6+12+13*6+2), dtype=np.float32)
  
    feats = [pool.apply_async(get_features, args=(load_header(header_files[i]),load_recording(recording_files[i]),two_leads)) for i in range(num_recordings)]
    for i in range(num_recordings):
        data[i,:] = feats[i].get() 
    del feats  
    gc.collect()                
    
    # Train 2-lead ECG model.
    print('Training 2-lead ECG model...')

    leads = two_leads

    top1 = [int(t) for t in top2]
    
#     feature_indices = [int(t) for t in range((12*2)+2+6+12+13*6+2)]
#     classifier = RandomForestClassifier(n_estimators=n_estimators, random_state=random_state,n_jobs=-1,bootstrap=True,max_features='sqrt',min_samples_split=min_samples_split,min_samples_leaf = min_samples_leaf).fit(SimpleImputer().fit(data[:, feature_indices]).transform(data[:, feature_indices]), labels)
#     top1 = [i for i in np.argsort(classifier.feature_importances_)[::-1][:100]]
#     print(top1)
#     del classifier
#     del feature_indices

    imputer = SimpleImputer().fit(data[:, [int(t) for t in top1]])
    classifier = RandomForestClassifier(n_estimators=n_estimators, max_leaf_nodes=max_leaf_nodes,n_jobs=-1, random_state=random_state).fit(imputer.transform(data[:, [int(t) for t in top1]]), labels)
    print('Saving 2-lead ECG model...')      
    save_model(model_directory, leads, classes, imputer, classifier)

    del top1
    del imputer 
    del classifier
    del data
    gc.collect()    
        
    
    
    print('Extracting 3-lead features...')    
    
    data = np.zeros((num_recordings, (12*3)+3+6+12+13*6+2), dtype=np.float32)
  
    #This does feature extraction in parallel and stores it in the variable features
    feats = [pool.apply_async(get_features, args=(load_header(header_files[i]),load_recording(recording_files[i]),three_leads)) for i in range(num_recordings)]
    for i in range(num_recordings):
        data[i,:] = feats[i].get() 
    del feats   
    gc.collect()        
       
    # Train 3-lead ECG model.
    print('Training 3-lead ECG model...')

    leads = three_leads
    
    top1 = [int(t) for t in top3]    

#     feature_indices = [int(t) for t in range((12*3)+3+6+12+13*6+2)]
#     classifier = RandomForestClassifier(n_estimators=n_estimators, random_state=random_state,n_jobs=-1,bootstrap=True,max_features='sqrt',min_samples_split=min_samples_split,min_samples_leaf = min_samples_leaf).fit(SimpleImputer().fit(data[:, feature_indices]).transform(data[:, feature_indices]), labels)
#     top1 = [i for i in np.argsort(classifier.feature_importances_)[::-1][:100]]
#     print(top1)
#     del classifier
#     del feature_indices

    imputer = SimpleImputer().fit(data[:, [int(t) for t in top1]])
    classifier = RandomForestClassifier(n_estimators=n_estimators, max_leaf_nodes=max_leaf_nodes, n_jobs=-1,random_state=random_state).fit(imputer.transform(data[:, [int(t) for t in top1]]), labels)
    print('Saving 3-lead ECG model...')      
    save_model(model_directory, leads, classes, imputer, classifier)
    del data
    del top1
    del imputer 
    del classifier
    gc.collect()    
    
    print('Extracting 4-lead features...')    
    
    data = np.zeros((num_recordings, (12*4)+4+6+12+13*6+2), dtype=np.float32)
  
    #This does feature extraction in parallel and stores it in the variable features
    feats = [pool.apply_async(get_features, args=(load_header(header_files[i]),load_recording(recording_files[i]),four_leads)) for i in range(num_recordings)]
    for i in range(num_recordings):
        data[i,:] = feats[i].get() 
    del feats  
    gc.collect()        
  
    # Train 4-lead ECG model.
    print('Training 4-lead ECG model...')
    
    leads = four_leads

    top1 = [int(t) for t in top4]    

#     feature_indices = [int(t) for t in range((12*4)+4+6+12+13*6+2)]
#     classifier = RandomForestClassifier(n_estimators=n_estimators, random_state=random_state,n_jobs=-1,bootstrap=True,max_features='sqrt',min_samples_split=min_samples_split,min_samples_leaf = min_samples_leaf).fit(SimpleImputer().fit(data[:, feature_indices]).transform(data[:, feature_indices]), labels)
#     top1 = [i for i in np.argsort(classifier.feature_importances_)[::-1][:100]]
#     print(top1)
#     del classifier
#     del feature_indices

    imputer = SimpleImputer().fit(data[:, [int(t) for t in top1]])
    classifier = RandomForestClassifier(n_estimators=n_estimators, max_leaf_nodes=max_leaf_nodes, n_jobs=-1,random_state=random_state).fit(imputer.transform(data[:, [int(t) for t in top1]]), labels)
    print('Saving 4-lead ECG model...')      
    save_model(model_directory, leads, classes, imputer, classifier)
    del data
    del top1
    del imputer 
    del classifier
    gc.collect()
    
    
    
    print('Extracting 6-lead features...')    
    
    data = np.zeros((num_recordings, (12*6)+6+6+12+13*6+2), dtype=np.float32)
  
    #This does feature extraction in parallel and stores it in the variable features
    feats = [pool.apply_async(get_features, args=(load_header(header_files[i]),load_recording(recording_files[i]),six_leads)) for i in range(num_recordings)]
    for i in range(num_recordings):
        data[i,:] = feats[i].get() 
    del feats  
    gc.collect()        
   
    # Train 6-lead ECG model.
    print('Training 6-lead ECG model...')
    
    leads = six_leads

    top1 = [int(t) for t in top6]    

#     feature_indices = [int(t) for t in range((12*6)+6+6+12+13*6+2)]
#     classifier = RandomForestClassifier(n_estimators=n_estimators, random_state=random_state,n_jobs=-1,bootstrap=True,max_features='sqrt',min_samples_split=min_samples_split,min_samples_leaf = min_samples_leaf).fit(SimpleImputer().fit(data[:, feature_indices]).transform(data[:, feature_indices]), labels)
#     top1 = [i for i in np.argsort(classifier.feature_importances_)[::-1][:100]]
#     print(top1)
#     del classifier
#     del feature_indices

    imputer = SimpleImputer().fit(data[:, [int(t) for t in top1]])
    classifier = RandomForestClassifier(n_estimators=n_estimators, max_leaf_nodes=max_leaf_nodes, n_jobs=-1,random_state=random_state).fit(imputer.transform(data[:, [int(t) for t in top1]]), labels)
    print('Saving 6-lead ECG model...')      
    save_model(model_directory, leads, classes, imputer, classifier)
    del data
    del top1
    del imputer 
    del classifier
    gc.collect()
       
    print('Extracting 12-lead features...')

    data = np.zeros((num_recordings, (12*12)+12+6+12+13*6+2), dtype=np.float32)
  
    #This does feature extraction in parallel and stores it in the variable features
    feats = [pool.apply_async(get_features, args=(load_header(header_files[i]),load_recording(recording_files[i]),twelve_leads)) for i in range(num_recordings)]
    for i in range(num_recordings):
        data[i,:] = feats[i].get() 
    del feats   
    gc.collect()    

    # Train 12-lead ECG model.
    print('Training 12-lead ECG model...')
    
    leads = twelve_leads

    top1 = [int(t) for t in top12] 

#     feature_indices = [int(t) for t in range((12*12)+12+6+12+13*6+2)]
#     classifier = RandomForestClassifier(n_estimators=n_estimators, random_state=random_state,n_jobs=-1,bootstrap=True,max_features='sqrt',min_samples_split=min_samples_split,min_samples_leaf = min_samples_leaf).fit(SimpleImputer().fit(data[:, feature_indices]).transform(data[:, feature_indices]), labels)
# #    print(np.sort(classifier.feature_importances_)[::-1][:50])
#     top1 = [i for i in np.argsort(classifier.feature_importances_)[::-1][:100]]
#     print(top1)
#     del classifier
#     del feature_indices

    imputer = SimpleImputer().fit(data[:, [int(t) for t in top1]])
    classifier = RandomForestClassifier(n_estimators=n_estimators, max_leaf_nodes=max_leaf_nodes, n_jobs=-1,random_state=random_state).fit(imputer.transform(data[:, [int(t) for t in top1]]), labels)
    
    print('Saving 12-lead ECG model...')  
    save_model(model_directory, leads, classes, imputer, classifier)
    
    pool.close()
    pool.join() 

    del data
    del top1
    del imputer 
    del classifier
    gc.collect()    
    
    


################################################################################
#
# Running trained model function
#
################################################################################

# Run your trained model. This function is *required*. You should edit this function to add your code, but do *not* change the arguments of this function.
def run_model(model, header, recording):
    classes = model['classes']
    leads = model['leads']
    imputer = model['imputer']
    classifier = model['classifier']
    classes = model['classes']
    leads = model['leads']
    imputer = model['imputer']
    classifier = model['classifier']
    
    top12 = [252, 121, 39, 125, 107, 123, 37, 47, 23, 127, 95, 119, 135, 147, 11, 43, 133, 3, 164, 41, 51, 9, 82, 45, 131, 145, 162, 137, 143, 94, 163, 59, 27, 139, 13, 106, 83, 15, 71, 80, 70, 92, 109, 35, 21, 141, 7, 113, 53, 57, 17, 22, 19, 115, 5, 73, 33, 55, 75, 1, 111, 49, 85, 69, 81, 63, 100, 30, 67, 155, 65, 101, 104, 249, 97, 93, 29, 90, 154, 31, 91, 118, 87, 34, 129, 153, 231, 99, 42, 77, 124, 114, 66, 103, 89, 61, 78, 16, 105, 126]

    top6 =  [174, 47, 23, 39, 11, 37, 75, 45, 41, 59, 43, 3, 9, 73, 71, 51, 15, 70, 27, 22, 21, 13, 34, 35, 19, 17, 57, 33, 7, 65, 86, 5, 153, 85, 68, 53, 69, 63, 67, 54, 55, 171, 4, 1, 29, 31, 42, 32, 20, 84, 49, 46, 18, 66, 40, 6, 61, 25, 72, 30, 28, 76, 64, 16, 52, 58, 99, 77, 150, 169, 60, 151, 44, 12, 168, 26, 36, 74, 24, 48, 62, 14, 56, 10, 38, 0, 50, 159, 8, 2, 157, 158, 156, 81, 83, 82, 170, 79, 80, 152]


    top4 =  [148, 23, 11, 47, 9, 15, 46, 3, 49, 13, 35, 17, 27, 21, 19, 44, 45, 60, 5, 41, 37, 7, 39, 33, 22, 43, 34, 1, 58, 42, 59, 29, 31, 25, 40, 18, 30, 16, 6, 28, 20, 48, 38, 4, 127, 32, 51, 36, 24, 145, 26, 12, 50, 73, 14, 0, 2, 10, 124, 125, 143, 142, 8, 56, 133, 55, 131, 130, 132, 54, 57, 52, 126, 139, 53, 68, 66, 62, 65, 137, 64, 136, 61, 123, 99, 67, 135, 111, 69, 134, 105, 63, 87, 110, 144, 93, 86, 116, 98, 122]

    top3 =  [135, 23, 35, 11, 3, 9, 15, 34, 37, 13, 21, 17, 19, 33, 32, 5, 27, 7, 22, 30, 29, 31, 25, 1, 18, 47, 28, 6, 45, 46, 26, 20, 36, 16, 12, 4, 114, 24, 38, 14, 132, 0, 10, 60, 112, 111, 2, 8, 130, 129, 43, 42, 120, 118, 119, 44, 55, 117, 126, 113, 41, 49, 40, 48, 54, 56, 122, 74, 39, 86, 121, 52, 50, 124, 98, 53, 84, 123, 69, 80, 110, 131, 115, 133, 73, 103, 104, 116, 97, 51, 82, 134, 68, 106, 128, 85, 91, 70, 109, 79]

    top2 =  [122, 11, 23, 3, 15, 9, 25, 13, 17, 21, 19, 22, 7, 33, 5, 18, 34, 16, 4, 1, 6, 20, 24, 12, 32, 14, 0, 95, 47, 2, 10, 119, 93, 8, 101, 117, 71, 69, 92, 116, 98, 99, 68, 83, 106, 30, 80, 107, 105, 41, 81, 31, 29, 35, 39, 104, 77, 59, 94, 75, 40, 73, 118, 36, 28, 57, 53, 38, 27, 67, 108, 56, 97, 96, 51, 37, 43, 60, 26, 109, 90, 113, 70, 85, 42, 74, 72, 66, 79, 91, 65, 82, 45, 121, 84, 111, 62, 78, 63, 115]



# Load features.
    num_leads = len(leads)
    if num_leads==12:
        top = top12
    elif num_leads==6:
        top = [int(t) for t in top6]
    elif num_leads==4:
        top = [int(t) for t in top4]    
    elif num_leads==3:
        top = [int(t) for t in top3]
    else:
        top = [int(t) for t in top2]

    data = get_features(header, recording, leads)
    feature_indices = [i for i in top]
    features = data[feature_indices]
   
    # Impute missing data.
    features = features.reshape(1, -1)
    features = imputer.transform(features)
    probabilities = classifier.predict_proba(features)
    for p in range(len(probabilities)):
        if len(probabilities[p][0])==1:
            probabilities[p] = np.array([[1, 0]])
    probabilities = np.asarray(probabilities, dtype=np.float32)[:, 0, 1]
    L = sorted(probabilities, reverse=True)
    labels = np.zeros(len(probabilities))
    if max(probabilities>=0.3):
        for i in range(len(labels)):    
            if probabilities[i]>=0.3:
                labels[i] = 1
    else:
        M = L[0]   
        for t in range(1,len(L)):
                if (L[t]/M)>=0.85 and L[t]>=0.13:
                    M = L[t]
        for i in range(len(labels)):    
            if probabilities[i]>=M:
                labels[i] = 1

    labels = np.asarray(labels, dtype=np.int)
    
    print(np.where(labels>0)[0])
   
    return classes, labels, probabilities


################################################################################
#
# File I/O functions
#
################################################################################

# Save a trained model. This function is not required. You can change or remove it.
def save_model(model_directory, leads, classes, imputer, classifier):
    d = {'leads': leads, 'classes': classes, 'imputer': imputer, 'classifier': classifier}
    filename = os.path.join(model_directory, get_model_filename(leads))
    joblib.dump(d, filename, protocol=0)

# Load a trained model. This function is *required*. You should edit this function to add your code, but do *not* change the arguments of this function.
def load_model(model_directory, leads):
    filename = os.path.join(model_directory, get_model_filename(leads))
    return joblib.load(filename)

# Define the filename(s) for the trained models. This function is not required. You can change or remove it.
def get_model_filename(leads):
    sorted_leads = sort_leads(leads)
    return 'model_' + '-'.join(sorted_leads) + '.sav'

################################################################################
#
# Feature extraction function
#
################################################################################

# Extract features from the header and recording. This function is not required. You can change or remove it.
def get_features(header, recording, LEADS):
    
    pca = PCA(n_components=3)
    
    # Extract age

    age = get_age(header)
    if age is None:
        age = float('nan')

    # Extract sex. Encode as 0 for female, 1 for male, and NaN for other.
    sex = get_sex(header)
    if sex in ('Female', 'female', 'F', 'f'):
        sex = 0
    elif sex in ('Male', 'male', 'M', 'm'):
        sex = 1
    else:
        sex = float('nan')
        
    
    Recording = np.copy(recording)
    numpoints=900
    skip = int(3)
    numslice = 6
    
    # Reorder/reselect leads in recordings.
    available_leads = get_leads(header)
    indices = list()           
     
    leads = LEADS
    if len(leads)==12:
        for lead in leads:
            i = available_leads.index(lead)
            indices.append(i)

        # Reorder/reselect leads in recordings.
        recording2 = choose_leads(recording, header, leads)

        # Pre-process recordings.
#        adc_gains = get_adc_gains(header, leads)
#        baselines = get_baselines(header, leads)

        num_leads = len(leads)
#        for i in range(num_leads):
#            recording2[i, :] = (recording2[i, :] - baselines[i]) / adc_gains[i]

        # Compute the root mean square of each ECG lead signal.
        rms = np.zeros(num_leads, dtype=np.float32)
        for i in range(num_leads):
            x = recording2[i, :]
            rms[i] = np.sqrt(abs(np.sum(x**2) / np.size(x)))

        RR = np.array([]) 
        for lead in leads:
            RR = np.append(RR,get_12ECG_features(Recording[indices, :],header,twelve_leads.index(lead)))
        All = np.append(RR,rms)
        #Extract TDA Features

        slice = 1
        a = np.array(np.arange(numpoints))
        a = a[a%skip==1]
        PTCLOUD = pca.fit_transform(StandardScaler().fit_transform(np.transpose(recording2[:,a])))
        maxdimension=1
        diagram1 = ripser(PTCLOUD, maxdim=maxdimension)['dgms']
        d1diagram1 = diagram1[1]
        diagram1 = diagram1[0][:,1][:-1]   
        denom = persim.persistent_entropy.persistent_entropy(d1diagram1)
#       diagram1 = diagram1[0]
        #d = dendrogram(linkage(PTCLOUD, 'single'),distance_sort='descending',no_plot=True)        
        #dtemp = np.sort(np.array([d['dcoord'][t][1] for t in range(len(d['dcoord'])) if t>0]))
        #diagram1 = np.array([[0,t] for t in dtemp])
        #diagram1 = np.append(diagram1,np.array([[0,np.inf]]))
        #diagram1 = diagram1.reshape(int(len(diagram1)/2),2)
        #diagram1 = np.append(diagram1,[[0,np.inf]])
        slice = slice + 1
        distances = np.array([])
        persentro = np.array([])
        persentrodiff = np.array([])
        perslands = np.array([])
        perslandstemp =  persland(d1diagram1) 
        while (slice*numpoints<len(recording2[1,:]) and slice<numslice):
            a = np.array(np.arange((slice-1)*numpoints,(slice)*numpoints))
            a = a[a%skip==1]
            PTCLOUD2 = pca.fit_transform(StandardScaler().fit_transform(np.transpose(recording2[:,a])))
            diagram2 = ripser(PTCLOUD2, maxdim=maxdimension)['dgms']
            d1diagram2 = diagram2[1]
            diagram2 = diagram2[0][:,1][:-1]  
            numerator = persim.persistent_entropy.persistent_entropy(d1diagram2)
            #d = dendrogram(linkage(PTCLOUD, 'single'),distance_sort='descending',no_plot=True)
            #dtemp = np.sort(np.array([d['dcoord'][t][1] for t in range(len(d['dcoord'])) if t>0]))
            #dtemp = dtemp/max(dtemp)
            #diagram2 = np.array([[0,t] for t in dtemp])
            #diagram2 = np.append(diagram2,np.array([[0,np.inf]]))
            #diagram2 = diagram2.reshape(int(len(diagram2)/2),2)
            distances = np.append(distances,bottleneck0(diagram1, diagram2))
            perslandtemp2 = persland(d1diagram2)

            perslands = np.append(perslands,np.array([perslandtemp2-perslandstemp]))
            persentro = np.append(persentro,np.copy(denom))
            persentrodiff = np.append(persentrodiff,abs(numerator-denom))
            #persentro0 = np.append(persentro,persim.persistent_entropy.persistent_entropy(diagram1))
            diagram1 = np.copy(diagram2)
            d1diagram1 = np.copy(d1diagram2)
            denom = np.copy(numerator)
            perslandstemp = np.copy(perslandtemp2)
            slice +=1
        if (max(distances)>0):
            distances  = distances/max(distances)
        All = np.append(All,[min(distances),max(distances)-min(distances),np.mean(distances)])
        All = np.append(All,[np.std(distances),skew(distances),kurtosis(distances)])
        perslands = perslands.reshape(int(len(perslands)/13),13)
        if len(persentro)==0:
            persentro = ([.001, 0.002, 0.001])        
        if len(persentrodiff)==0:
            persentrodiff = ([.001, 0.002, 0.001])    
        if len(perslands)==0:
            perslands = ([.001, 0.002, 0.001])        
        All = np.append(All,[min(persentro),max(persentro),np.mean(persentro)])
        All = np.append(All,[np.std(persentro),skew(persentro),kurtosis(persentro)])
        All = np.append(All,[min(persentrodiff),max(persentrodiff),np.mean(persentrodiff)])
        All = np.append(All,[np.std(persentrodiff),skew(persentrodiff),kurtosis(persentrodiff)])
        for t in range(13):
            All = np.append(All,[min(perslands[:,t]),max(perslands[:,t]),np.mean(perslands[:,t])])
            All = np.append(All,[np.std(perslands[:,t]),skew(perslands[:,t]),kurtosis(perslands[:,t])])
        All = np.append(All,[age,sex])

    elif len(leads)==6:
        for lead in leads:
            i = available_leads.index(lead)
            indices.append(i)

        # Reorder/reselect leads in recordings.
        recording2 = choose_leads(recording, header, leads)

        # Pre-process recordings.
#        adc_gains = get_adc_gains(header, leads)
#        baselines = get_baselines(header, leads)

        num_leads = len(leads)
#        for i in range(num_leads):
#            recording2[i, :] = (recording2[i, :] - baselines[i]) / adc_gains[i]

        # Compute the root mean square of each ECG lead signal.
        rms = np.zeros(num_leads, dtype=np.float32)
        for i in range(num_leads):
            x = recording2[i, :]
            rms[i] = np.sqrt(abs(np.sum(x**2) / np.size(x)))

        RR = np.array([]) 
        for lead in leads:
            RR = np.append(RR,get_12ECG_features(Recording[indices, :],header,six_leads.index(lead)))
        All = np.append(RR,rms)

        slice = 1
        a = np.array(np.arange(numpoints))
        a = a[a%skip==1]
        PTCLOUD = pca.fit_transform(StandardScaler().fit_transform(np.transpose(recording2[:,a])))

        maxdimension=1
        diagram1 = ripser(PTCLOUD, maxdim=maxdimension)['dgms']
        d1diagram1 = diagram1[1]
        diagram1 = diagram1[0][:,1][:-1]   
        denom = persim.persistent_entropy.persistent_entropy(d1diagram1)
#       diagram1 = diagram1[0]
        #d = dendrogram(linkage(PTCLOUD, 'single'),distance_sort='descending',no_plot=True)        
        #dtemp = np.sort(np.array([d['dcoord'][t][1] for t in range(len(d['dcoord'])) if t>0]))
        #diagram1 = np.array([[0,t] for t in dtemp])
        #diagram1 = np.append(diagram1,np.array([[0,np.inf]]))
        #diagram1 = diagram1.reshape(int(len(diagram1)/2),2)
        #diagram1 = np.append(diagram1,[[0,np.inf]])
        slice = slice + 1
        distances = np.array([])
        persentro = np.array([])
        persentrodiff = np.array([])
        perslands = np.array([])
        perslandstemp =  persland(d1diagram1) 
        while (slice*numpoints<len(recording2[1,:]) and slice<numslice):
            a = np.array(np.arange((slice-1)*numpoints,(slice)*numpoints))
            a = a[a%skip==1]
            PTCLOUD2 = pca.fit_transform(StandardScaler().fit_transform(np.transpose(recording2[:,a])))
            diagram2 = ripser(PTCLOUD2, maxdim=maxdimension)['dgms']
            d1diagram2 = diagram2[1]
            diagram2 = diagram2[0][:,1][:-1]  
            numerator = persim.persistent_entropy.persistent_entropy(d1diagram2)
            #d = dendrogram(linkage(PTCLOUD, 'single'),distance_sort='descending',no_plot=True)
            #dtemp = np.sort(np.array([d['dcoord'][t][1] for t in range(len(d['dcoord'])) if t>0]))
            #dtemp = dtemp/max(dtemp)
            #diagram2 = np.array([[0,t] for t in dtemp])
            #diagram2 = np.append(diagram2,np.array([[0,np.inf]]))
            #diagram2 = diagram2.reshape(int(len(diagram2)/2),2)
            distances = np.append(distances,bottleneck0(diagram1, diagram2))
            perslandtemp2 = persland(d1diagram2)
            perslands = np.append(perslands,np.array([perslandtemp2-perslandstemp]))
            persentro = np.append(persentro,np.copy(denom))
            persentrodiff = np.append(persentrodiff,abs(numerator-denom))
            #persentro0 = np.append(persentro,persim.persistent_entropy.persistent_entropy(diagram1))
            diagram1 = np.copy(diagram2)
            d1diagram1 = np.copy(d1diagram2)
            denom = np.copy(numerator)
            perslandstemp = np.copy(perslandtemp2)
            slice +=1
        if (max(distances)>0):
            distances  = distances/max(distances)
        All = np.append(All,[min(distances),max(distances)-min(distances),np.mean(distances)])
        All = np.append(All,[np.std(distances),skew(distances),kurtosis(distances)])
        perslands = perslands.reshape(int(len(perslands)/13),13)
        if len(persentro)==0:
            persentro = ([.001, 0.002, 0.001])        
        if len(persentrodiff)==0:
            persentrodiff = ([.001, 0.002, 0.001])    
        if len(perslands)==0:
            persentro = ([.001, 0.002, 0.001])        
        All = np.append(All,[min(persentro),max(persentro),np.mean(persentro)])
        All = np.append(All,[np.std(persentro),skew(persentro),kurtosis(persentro)])
        All = np.append(All,[min(persentrodiff),max(persentrodiff),np.mean(persentrodiff)])
        All = np.append(All,[np.std(persentrodiff),skew(persentrodiff),kurtosis(persentrodiff)])
        for t in range(13):
            All = np.append(All,[min(perslands[:,t]),max(perslands[:,t]),np.mean(perslands[:,t])])
            All = np.append(All,[np.std(perslands[:,t]),skew(perslands[:,t]),kurtosis(perslands[:,t])])
        All = np.append(All,[age,sex])
         
    elif len(leads)==4:
        for lead in leads:
            i = available_leads.index(lead)
            indices.append(i)

        # Reorder/reselect leads in recordings.
        recording2 = choose_leads(recording, header, leads)

        # Pre-process recordings.
#        adc_gains = get_adc_gains(header, leads)
#        baselines = get_baselines(header, leads)

        num_leads = len(leads)
#        for i in range(num_leads):
#            recording2[i, :] = (recording2[i, :] - baselines[i]) / adc_gains[i]

        # Compute the root mean square of each ECG lead signal.
        rms = np.zeros(num_leads, dtype=np.float32)
        for i in range(num_leads):
            x = recording2[i, :]
            rms[i] = np.sqrt(abs(np.sum(x**2) / np.size(x)))
        RR = np.array([]) 
        for lead in leads:
            RR = np.append(RR,get_12ECG_features(Recording[indices, :],header,four_leads.index(lead)))
        All = np.append(RR,rms)
        #Extract TDA Features
        slice = 1
        a = np.array(np.arange(numpoints))
        a = a[a%skip==1]
        PTCLOUD = pca.fit_transform(StandardScaler().fit_transform(np.transpose(recording2[:,a])))

        maxdimension=1
        diagram1 = ripser(PTCLOUD, maxdim=maxdimension)['dgms']
        d1diagram1 = diagram1[1]
        diagram1 = diagram1[0][:,1][:-1]   
        denom = persim.persistent_entropy.persistent_entropy(d1diagram1)
#       diagram1 = diagram1[0]
        #d = dendrogram(linkage(PTCLOUD, 'single'),distance_sort='descending',no_plot=True)        
        #dtemp = np.sort(np.array([d['dcoord'][t][1] for t in range(len(d['dcoord'])) if t>0]))
        #diagram1 = np.array([[0,t] for t in dtemp])
        #diagram1 = np.append(diagram1,np.array([[0,np.inf]]))
        #diagram1 = diagram1.reshape(int(len(diagram1)/2),2)
        #diagram1 = np.append(diagram1,[[0,np.inf]])
        slice = slice + 1
        distances = np.array([])
        persentro = np.array([])
        persentrodiff = np.array([])
        perslands = np.array([])
        perslandstemp =  persland(d1diagram1) 
        while (slice*numpoints<len(recording2[1,:]) and slice<numslice):
            a = np.array(np.arange((slice-1)*numpoints,(slice)*numpoints))
            a = a[a%skip==1]
            PTCLOUD2 = pca.fit_transform(StandardScaler().fit_transform(np.transpose(recording2[:,a])))
            diagram2 = ripser(PTCLOUD2, maxdim=maxdimension)['dgms']
            d1diagram2 = diagram2[1]
            diagram2 = diagram2[0][:,1][:-1]  
            numerator = persim.persistent_entropy.persistent_entropy(d1diagram2)
            #d = dendrogram(linkage(PTCLOUD, 'single'),distance_sort='descending',no_plot=True)
            #dtemp = np.sort(np.array([d['dcoord'][t][1] for t in range(len(d['dcoord'])) if t>0]))
            #dtemp = dtemp/max(dtemp)
            #diagram2 = np.array([[0,t] for t in dtemp])
            #diagram2 = np.append(diagram2,np.array([[0,np.inf]]))
            #diagram2 = diagram2.reshape(int(len(diagram2)/2),2)
            distances = np.append(distances,bottleneck0(diagram1, diagram2))
            perslandtemp2 = persland(d1diagram2)
            perslands = np.append(perslands,np.array([perslandtemp2-perslandstemp]))
            persentro = np.append(persentro,np.copy(denom))
            persentrodiff = np.append(persentrodiff,abs(numerator-denom))
            #persentro0 = np.append(persentro,persim.persistent_entropy.persistent_entropy(diagram1))
            diagram1 = np.copy(diagram2)
            d1diagram1 = np.copy(d1diagram2)
            denom = np.copy(numerator)
            perslandstemp = np.copy(perslandtemp2)
            slice +=1
        if (max(distances)>0):
            distances  = distances/max(distances)
        All = np.append(All,[min(distances),max(distances)-min(distances),np.mean(distances)])
        All = np.append(All,[np.std(distances),skew(distances),kurtosis(distances)])
        perslands = perslands.reshape(int(len(perslands)/13),13)
        if len(persentro)==0:
            persentro = ([.001, 0.002, 0.001])        
        if len(persentrodiff)==0:
            persentrodiff = ([.001, 0.002, 0.001])    
        if len(perslands)==0:
            persentro = ([.001, 0.002, 0.001])         
        All = np.append(All,[min(persentro),max(persentro),np.mean(persentro)])
        All = np.append(All,[np.std(persentro),skew(persentro),kurtosis(persentro)])
        All = np.append(All,[min(persentrodiff),max(persentrodiff),np.mean(persentrodiff)])
        All = np.append(All,[np.std(persentrodiff),skew(persentrodiff),kurtosis(persentrodiff)])
        for t in range(13):
            All = np.append(All,[min(perslands[:,t]),max(perslands[:,t]),np.mean(perslands[:,t])])
            All = np.append(All,[np.std(perslands[:,t]),skew(perslands[:,t]),kurtosis(perslands[:,t])])
        All = np.append(All,[age,sex])
    #else:
    #    All = np.append(All,[np.zeros(8+4+(12*4))])

    #if len(available_leads)==12 or len(available_leads)==3:    
    elif len(leads)==3:
        for lead in leads:
            i = available_leads.index(lead)
            indices.append(i)

        # Reorder/reselect leads in recordings.
        recording2 = choose_leads(recording, header, leads)

        # Pre-process recordings.
#        adc_gains = get_adc_gains(header, leads)
#        baselines = get_baselines(header, leads)

        num_leads = len(leads)
#        for i in range(num_leads):
#            recording2[i, :] = (recording2[i, :] - baselines[i]) / adc_gains[i]

        # Compute the root mean square of each ECG lead signal.
        rms = np.zeros(num_leads, dtype=np.float32)
        for i in range(num_leads):
            x = recording2[i, :]
            rms[i] = np.sqrt(abs(np.sum(x**2) / np.size(x)))
        RR = np.array([]) 
        for lead in leads:
            RR = np.append(RR,get_12ECG_features(Recording[indices, :],header,three_leads.index(lead)))
        All = np.append(RR,rms)
        #Extract TDA Features
        slice = 1
        a = np.array(np.arange(numpoints))
        a = a[a%skip==1]
        PTCLOUD = StandardScaler().fit_transform(np.transpose(recording2[:,a]))

        maxdimension=1
        diagram1 = ripser(PTCLOUD, maxdim=maxdimension)['dgms']
        d1diagram1 = diagram1[1]
        diagram1 = diagram1[0][:,1][:-1]   
        denom = persim.persistent_entropy.persistent_entropy(d1diagram1)
#       diagram1 = diagram1[0]
        #d = dendrogram(linkage(PTCLOUD, 'single'),distance_sort='descending',no_plot=True)        
        #dtemp = np.sort(np.array([d['dcoord'][t][1] for t in range(len(d['dcoord'])) if t>0]))
        #diagram1 = np.array([[0,t] for t in dtemp])
        #diagram1 = np.append(diagram1,np.array([[0,np.inf]]))
        #diagram1 = diagram1.reshape(int(len(diagram1)/2),2)
        #diagram1 = np.append(diagram1,[[0,np.inf]])
        slice = slice + 1
        distances = np.array([])
        persentro = np.array([])
        persentrodiff = np.array([])
        perslands = np.array([])
        perslandstemp =  persland(d1diagram1) 
        while (slice*numpoints<len(recording2[1,:]) and slice<numslice):
            a = np.array(np.arange((slice-1)*numpoints,(slice)*numpoints))
            a = a[a%skip==1]
            PTCLOUD2 = StandardScaler().fit_transform(np.transpose(recording2[:,a]))
            diagram2 = ripser(PTCLOUD2, maxdim=maxdimension)['dgms']
            d1diagram2 = diagram2[1]
            diagram2 = diagram2[0][:,1][:-1]  
            numerator = persim.persistent_entropy.persistent_entropy(d1diagram2)
            #d = dendrogram(linkage(PTCLOUD, 'single'),distance_sort='descending',no_plot=True)
            #dtemp = np.sort(np.array([d['dcoord'][t][1] for t in range(len(d['dcoord'])) if t>0]))
            #dtemp = dtemp/max(dtemp)
            #diagram2 = np.array([[0,t] for t in dtemp])
            #diagram2 = np.append(diagram2,np.array([[0,np.inf]]))
            #diagram2 = diagram2.reshape(int(len(diagram2)/2),2)
            distances = np.append(distances,bottleneck0(diagram1, diagram2))
            perslandtemp2 = persland(d1diagram2)
            perslands = np.append(perslands,np.array([perslandtemp2-perslandstemp]))
            persentro = np.append(persentro,np.copy(denom))
            persentrodiff = np.append(persentrodiff,abs(numerator-denom))
            #persentro0 = np.append(persentro,persim.persistent_entropy.persistent_entropy(diagram1))
            diagram1 = np.copy(diagram2)
            d1diagram1 = np.copy(d1diagram2)
            denom = np.copy(numerator)
            perslandstemp = np.copy(perslandtemp2)
            slice +=1
        if (max(distances)>0):
            distances  = distances/max(distances)
        All = np.append(All,[min(distances),max(distances)-min(distances),np.mean(distances)])
        All = np.append(All,[np.std(distances),skew(distances),kurtosis(distances)])
        perslands = perslands.reshape(int(len(perslands)/13),13)
        if len(persentro)==0:
            persentro = ([.001, 0.002, 0.001])        
        if len(persentrodiff)==0:
            persentrodiff = ([.001, 0.002, 0.001])    
        if len(perslands)==0:
            persentro = ([.001, 0.002, 0.001])         
        All = np.append(All,[min(persentro),max(persentro),np.mean(persentro)])
        All = np.append(All,[np.std(persentro),skew(persentro),kurtosis(persentro)])
        All = np.append(All,[min(persentrodiff),max(persentrodiff),np.mean(persentrodiff)])
        All = np.append(All,[np.std(persentrodiff),skew(persentrodiff),kurtosis(persentrodiff)])
        for t in range(13):
            All = np.append(All,[min(perslands[:,t]),max(perslands[:,t]),np.mean(perslands[:,t])])
            All = np.append(All,[np.std(perslands[:,t]),skew(perslands[:,t]),kurtosis(perslands[:,t])])
        All = np.append(All,[age,sex])
        
    #else:
    #    All = np.append(All,[np.zeros(8+3+(12*3))])

    #if len(available_leads)==12 or len(available_leads)==2:       
    elif len(leads)==2:
        for lead in leads:
            i = available_leads.index(lead)
            indices.append(i)

        # Reorder/reselect leads in recordings.
        recording2 = choose_leads(recording, header, leads)

        # Pre-process recordings.
#        adc_gains = get_adc_gains(header, leads)
#        baselines = get_baselines(header, leads)

        num_leads = len(leads)
#        for i in range(num_leads):
#            recording2[i, :] = (recording2[i, :] - baselines[i]) / adc_gains[i]

        # Compute the root mean square of each ECG lead signal.
        rms = np.zeros(num_leads, dtype=np.float32)
        for i in range(num_leads):
            x = recording2[i, :]
            rms[i] = np.sqrt(abs(np.sum(x**2) / np.size(x)))
        RR = np.array([]) 
        for lead in leads:
            RR = np.append(RR,get_12ECG_features(Recording[indices, :],header,two_leads.index(lead)))
        All = np.append(RR,rms)
        #Extract TDA Features
        slice = 1
        a = np.array(np.arange(numpoints))
        a = a[a%skip==1]
        PTCLOUD = StandardScaler().fit_transform(np.transpose(recording2[:,a]))

        maxdimension=1
        diagram1 = ripser(PTCLOUD, maxdim=maxdimension)['dgms']
        d1diagram1 = diagram1[1]
        diagram1 = diagram1[0][:,1][:-1]   
        denom = persim.persistent_entropy.persistent_entropy(d1diagram1)
#       diagram1 = diagram1[0]
        #d = dendrogram(linkage(PTCLOUD, 'single'),distance_sort='descending',no_plot=True)        
        #dtemp = np.sort(np.array([d['dcoord'][t][1] for t in range(len(d['dcoord'])) if t>0]))
        #diagram1 = np.array([[0,t] for t in dtemp])
        #diagram1 = np.append(diagram1,np.array([[0,np.inf]]))
        #diagram1 = diagram1.reshape(int(len(diagram1)/2),2)
        #diagram1 = np.append(diagram1,[[0,np.inf]])
        slice = slice + 1
        distances = np.array([])
        persentro = np.array([])
        persentrodiff = np.array([])
        perslands = np.array([])
        perslandstemp =  persland(d1diagram1) 
        while (slice*numpoints<len(recording2[1,:]) and slice<numslice):
            a = np.array(np.arange((slice-1)*numpoints,(slice)*numpoints))
            a = a[a%skip==1]
            PTCLOUD2 = StandardScaler().fit_transform(np.transpose(recording2[:,a]))
            diagram2 = ripser(PTCLOUD2, maxdim=maxdimension)['dgms']
            d1diagram2 = diagram2[1]
            diagram2 = diagram2[0][:,1][:-1]  
            numerator = persim.persistent_entropy.persistent_entropy(d1diagram2)
            #d = dendrogram(linkage(PTCLOUD, 'single'),distance_sort='descending',no_plot=True)
            #dtemp = np.sort(np.array([d['dcoord'][t][1] for t in range(len(d['dcoord'])) if t>0]))
            #dtemp = dtemp/max(dtemp)
            #diagram2 = np.array([[0,t] for t in dtemp])
            #diagram2 = np.append(diagram2,np.array([[0,np.inf]]))
            #diagram2 = diagram2.reshape(int(len(diagram2)/2),2)
            distances = np.append(distances,bottleneck0(diagram1, diagram2))
            perslandtemp2 = persland(d1diagram2)
            perslands = np.append(perslands,np.array([perslandtemp2-perslandstemp]))
            persentro = np.append(persentro,np.copy(denom))
            persentrodiff = np.append(persentrodiff,abs(numerator-denom))
            #persentro0 = np.append(persentro,persim.persistent_entropy.persistent_entropy(diagram1))
            diagram1 = np.copy(diagram2)
            d1diagram1 = np.copy(d1diagram2)
            denom = np.copy(numerator)
            perslandstemp = np.copy(perslandtemp2)
            slice +=1
        if (max(distances)>0):
            distances  = distances/max(distances)
        All = np.append(All,[min(distances),max(distances)-min(distances),np.mean(distances)])
        All = np.append(All,[np.std(distances),skew(distances),kurtosis(distances)])
        perslands = perslands.reshape(int(len(perslands)/13),13)
        if len(persentro)==0:
            persentro = ([.001, 0.002, 0.001])        
        if len(persentrodiff)==0:
            persentrodiff = ([.001, 0.002, 0.001])    
        if len(perslands)==0:
            persentro = ([.001, 0.002, 0.001])        
        All = np.append(All,[min(persentro),max(persentro),np.mean(persentro)])
        All = np.append(All,[np.std(persentro),skew(persentro),kurtosis(persentro)])
        All = np.append(All,[min(persentrodiff),max(persentrodiff),np.mean(persentrodiff)])
        All = np.append(All,[np.std(persentrodiff),skew(persentrodiff),kurtosis(persentrodiff)])
        for t in range(13):
            All = np.append(All,[min(perslands[:,t]),max(perslands[:,t]),np.mean(perslands[:,t])])
            All = np.append(All,[np.std(perslands[:,t]),skew(perslands[:,t]),kurtosis(perslands[:,t])])
        All = np.append(All,[age,sex])
  
    #else:
    #    All = np.append(All,[np.zeros(8+2+(12*2))])
    #print(All)
    return All


# Load challenge data.
def load_challenge_data(header_file):
    with open(header_file, 'r') as f:
        header = f.readlines()
    mat_file = header_file.replace('.hea', '.mat')
    x = loadmat(mat_file)
    recording = np.asarray(x['val'], dtype=np.float64)
    return recording, header


def bottleneck0(X,Y):
    """
    Returns the dimension 0 bottleneck distance between two non-empty persistence diagrams.
    All components are assumed to be born at the beginning of the filtration, 
    that is, all non-trivial points lie in the vertical axis.   
    
    Parameters
    ----------
    X : 1D numpy array
        Non-empty array of death times for persistence diagram 1.
    Y : 1D numpy array
        Non-empty array of death times for persistence diagram 2.
           
    Returns
    -------
    d : float
        The bottleneck distance between X and Y.

    Authors
    --------
    Paul Samuel P. Ignacio,
    University of the Philippines Baguio,
    Governor Pack Road, Baguio City 2600, Philippines
    
    Jay-Anne B. Bulauan,
    University of the Philippines Baguio,
    Governor Pack Road, Baguio City 2600, Philippines

    David Uminsky,
    University of Chicago, 5801 S Ellis Ave, Chicago,IL 60637, United States
        
    Examples
    --------
    >>> num_points1 = 100
    >>> num_points2 = 200
    >>> X = np.random.uniform(low=0.0, high=1000.0, size=num_points1)
    >>> Y = np.random.uniform(low=0.0, high=1000.0, size=num_points2)
    >>> bottleneck0(X,Y)
   
      
    """
    
    #Swap if X is longer
    if len(Y) < len(X):
        X_copy = np.copy(X)
        X = Y
        Y = X_copy  

    #Initialize
    X=-np.sort(-X) 
    Y=-np.sort(-Y)
    d=0
    N = len(X)
    Z = abs(X-Y[0:N])
    if Z.size==0 and len(Y)>N:
        return 0.5*Y[N]
    elif Z.size==0 and len(Y)==N:
        return 0
    else:
        l = np.argmax(Z) 
        dtemp = Z[l]

    # Maintain an initial bijection. Systematically modify this bijection to optimized
    # the norm between matched points until the bottleneck matching is recovered.
    if N != len(Y) and dtemp < 0.5*Y[N]:
        d = 0.5*Y[N]      
    elif l >= 0 and len(Z) > 1:
        while len(Z) > 1:
            k = 0.5*max(X[l],Y[l])
            if np.max(np.delete(Z,l)) <  k and k < dtemp: 
                d = k
                break

            elif np.max(np.delete(Z,l)) >= k:
                if len(Z[Z>=k]) == len(np.where(np.where(Z >= k)[0] >= l)[0]): 
                    d = k
                    break
                else:
                    Z=Z[0:l]
                    X=X[0:l]
                    Y=Y[0:l]
                    l=np.argmax(Z)
                    dtemp = Z[l]             
                    if len(Z) == 1:
                        d = min(dtemp,0.5*max(X[l],Y[l]))
                        break
            else:
                d = dtemp
                break
    else:
        # If one of the persistence diagrams is a singleton 
        d = min(dtemp,0.5*max(X[0],Y[0]))
   
    return d


#This is a new TDA feature
def persland(diagram):
    diagram=sorted(diagram, key=lambda t: (t[0], -t[1]))
    diagram=np.array(diagram)
    k = 0
    L = []
    Peaks = []
    Valleys = []
    while len(diagram) != 0:
        i = 0
        p = 0
        pair = diagram[i]
        diagram = np.delete(diagram, i, axis = 0)
        L.append(np.array([[-math.inf,0], [pair[0],0], [0.5*(pair[0] + pair[1]), 0.5*(pair[1] - pair[0])]]))
        Peaks.append(np.array([[0.5*(pair[0] + pair[1]), 0.5*(pair[1] - pair[0])]]))
        Valleys.append(np.array([[pair[0],0]]))
        while all(L[k][-1] != np.array([math.inf, 0])):
            if len(diagram) == 0 or p==len(diagram):
                L[k] = np.append(L[k], np.array([[pair[1], 0]]), axis=0)
                #L[k] = np.append(L[k], np.array([[math.inf, 0]]), axis=0)
                break
            ran = range(p, len(diagram))
            if pair[1] >= max(diagram[:,1][ran]):
                L[k] = np.append(L[k], np.array([[pair[1], 0]]), axis=0)
                #L[k] = np.append(L[k], np.array([[math.inf, 0]]), axis=0)
            else:
                u = np.where(diagram[:,1] > pair[1])[0]
                j = min(u[np.where(u >= p)[0]])
                pair1 = diagram[j]
                diagram = np.delete(diagram, j, axis = 0)
                p = j
                if pair1[0] > pair[1]:
                    L[k] = np.append(L[k], np.array([[pair[1], 0]]), axis = 0)
                    Valleys[k] = np.append(Valleys[k], np.array([[pair[1],0], [pair1[0],0]]), axis = 0)
                if pair1[0] >= pair[1]:
                    L[k] = np.append(L[k], np.array([[pair1[0], 0]]), axis = 0)
                else:
                    b_1 = 0.5*(pair1[0] + pair[1])
                    d_1 = 0.5*(pair[1] - pair1[0])
                    L[k] = np.append(L[k], np.array([[b_1, d_1]]), axis=0)
                    diagram = np.append(diagram, np.array([[pair1[0],pair[1]]]), axis = 0)
                    ran = range(p, len(diagram))
                    diagram[ran] = sorted(diagram[ran], key=lambda t: (t[0], -t[1]))
                    Valleys[k] = np.append(Valleys[k], np.array([[b_1,d_1]]), axis = 0)
                b_2 = 0.5*(pair1[0] + pair1[1])
                d_2 = 0.5*(pair1[1] - pair1[0])
                L[k] = np.append(L[k], np.array([[b_2,d_2]]), axis = 0)
                Peaks[k] = np.append(Peaks[k], np.array([[b_2,d_2]]), axis = 0)
                pair = pair1
        Valleys[k] = np.append(Valleys[k], np.array([[pair[1], 0]]), axis = 0)
        k+=1
    #Features from the Persistence Landscapes
    
    Order = k #Maximum Number of Layers
    Num_Peaks = np.zeros(k)
    Num_Valls = np.zeros(k)

    for i in range(k):
        Num_Peaks[i] = len(Peaks[i]) #Number of Peaks Per Layer
        Num_Valls[i] = len(Valleys[i]) #Number of Valleys Per Layer
    if len(Num_Peaks)>1:
        Mean_Peaks = np.mean(Num_Peaks) #Mean of Peaks
        STD_Peaks = np.std(Num_Peaks) #Standard Deviation of Peaks
        Skew_Peaks = skew(np.asarray(Num_Peaks)) #Skewness of Peaks
        Kurt_Peaks = kurtosis(np.asarray(Num_Peaks)) #Kurtosis of Peaks
    else:
        Mean_Peaks = 0
        STD_Peaks = 0
        Skew_Peaks = 0
        Kurt_Peaks = 0
    if len(Num_Valls)>1:
        Mean_Valls = np.mean(Num_Valls) #Mean of Valleys     
        STD_Valls = np.std(Num_Valls) #Standard Deviation of Valleys
        Skew_Valls = skew(np.asarray(Num_Valls)) #Skewness Valleys
        Kurt_Valls = kurtosis(np.asarray(Num_Valls)) #Kurtosis of Peaks
    else:
        Mean_Valls = 0
        STD_Valls = 0
        Skew_Valls = 0
        Kurt_Valls = 0
    
    Areas = np.zeros(k)
    if k>1:
        Rel_Areas = np.zeros(k-1)
        for i in range(k):
            q = L[i]
            for j in range(1,len(q) - 1):
                b = 0.5*(q[j][1] + q[j+1][1])*(q[j+1][0] - q[j][0])
                Areas[i] = Areas[i] + b
        for i in range(0,k-2):
            Rel_Areas[i] = (Areas[i+1]-Areas[i+2])/Areas[1]
        Rel_Areas[k-2]=Areas[k-1]/Areas[1]
    else:
        Rel_Areas = np.array([1])
    
    Mean_Areas = np.mean(Rel_Areas)    
    STD_Areas = np.std(Rel_Areas)
    Skew_Areas = skew(np.asarray(Rel_Areas))
    Kurt_Areas = kurtosis(np.asarray(Rel_Areas)) 
    
    ECGFeat_Land = np.array([Order, Mean_Peaks, STD_Peaks, Skew_Peaks, Kurt_Peaks, Mean_Valls, STD_Valls, Skew_Valls, Kurt_Valls, Mean_Areas, STD_Areas, Skew_Areas, Kurt_Areas])
    
    return ECGFeat_Land
